import logging

import regex

MP4UPLOAD_REGEX = regex.compile(r"player\|(.*)\|videojs")


def extract_480(splitted_values):
    return {
        "stream_url": "{4}://{18}.mp4upload.{1}:{70}/d/{69}/{68}.{67}".format(
            *splitted_values
        )
    }


def extract_any(splitted_values):
    return {
        "stream_url": "{3}://{18}.mp4upload.{0}:{73}/d/{72}/{71}.{70}".format(
            *splitted_values
        )
    }


def extract(session, url, **opts):
    logger = logging.getLogger("mp4upload-extractor")

    mp4upload_embed_page = session.get(url)
    if mp4upload_embed_page.text == "File was deleted":
        return []

    content = MP4UPLOAD_REGEX.search(mp4upload_embed_page.text).group(1).split("|")

    try:
        return [
            {
                **(extract_480 if "480" in content else extract_any)(content),
                "headers": {"referer": url, "ssl_verification": False},
            }
        ]
    except Exception as e:
        return logger.error("'%s' occurred when extracting from '%s'." % (e, url)) or []


extract.disabled = True



import time
from collections import defaultdict
from datetime import datetime

import click

from ...__version__ import __core__
from ...config import ANICHART, DATE_FORMAT, TIME_FORMAT
from .. import helpers
from ..http_client import client

gql = """query (
        $weekStart: Int,
        $weekEnd: Int,
        $page: Int,
){
    Page(page: $page) {
        pageInfo {
                hasNextPage
                total
        }
        airingSchedules(
                airingAt_greater: $weekStart
                airingAt_lesser: $weekEnd
        ) {
            id
            episode
            airingAt
            media {
                title {
                    userPreferred
                }
            }
        }
    }
}"""


def arrange_template(data):
    template = defaultdict(lambda: defaultdict(list))

    datetime_now = datetime.now()

    for airing in data[::-1]:
        datetime_object = datetime.fromtimestamp(airing.get("airingAt", 0))

        if datetime_object < datetime_now:
            date_string = f"Aired @ {datetime_object.strftime(DATE_FORMAT)}"
        else:
            date_string = f"On {datetime_object.strftime(DATE_FORMAT)}"

        time_string = datetime_object.strftime(TIME_FORMAT)

        template[date_string][time_string, datetime_object].append(
            {
                "name": airing.get("media", {}).get("title", {}).get("userPreferred"),
                "episode": airing.get("episode", 0),
            }
        )
    return template


def iter_schedules(session, unix_time):
    page = 1

    data = {}

    week_end = unix_time + 24 * 7 * 60 * 60

    while data.get("pageInfo", {}).get("hasNextPage", True):
        schedule_data = session.post(
            ANICHART,
            json={
                "query": gql,
                "variables": {
                    "weekStart": unix_time,
                    "weekEnd": week_end,
                    "page": page,
                },
            },
        )

        data = schedule_data.json().get("data", {}).get("Page", {})
        page += 1

        yield from data.get("airingSchedules", [])


@click.command(name="schedule", help="Know which animes are going over the air when.")
@click.option(
    "--offset",
    "-o",
    type=int,
    default=86400,
    help="Subtract the offset from the current time to get the schedule for aired anime.",
)
@helpers.decorators.logging_options()
@helpers.decorators.setup_loggers()
@helpers.decorators.banner_gift_wrapper(client, __core__)
def animdl_schedule(offset: int, **kwargs):

    including_yesterday = int(time.time()) - offset

    for date_format, child_component in arrange_template(
        list(iter_schedules(client, including_yesterday))
    ).items():
        click.secho(date_format, fg="cyan")
        for (time_format, _), anime_components in sorted(
            child_component.items(), key=lambda component: component[0][1], reverse=True
        ):
            click.echo(
                f"\t{click.style(time_format, fg='magenta')} - {{}}".format(
                    ", ".join(
                        f"{anime['name']} [{click.style(anime['episode'], fg='blue')}]"
                        for anime in anime_components
                    ),
                )
            )



import json
from functools import lru_cache

import regex
import yarl

from .utils import decipher_salted_aes

CONTENT_ID_REGEX = regex.compile(r"embed-6/([^?#&/.]+)")

SID_ENDPOINT = "https://api.enime.moe/tool/rapid-cloud/server-id"
SALT_SECRET_ENDPOINT = (
    "https://raw.githubusercontent.com/consumet/rapidclown/main/key.txt"
)


@lru_cache()
def get_associative_key(session, endpoint):
    return session.get(endpoint).text


def extract(session, url, **opts):

    sid = get_associative_key(session, SID_ENDPOINT)

    url = yarl.URL(url)

    ajax_response = session.get(
        f"https://{url.host}/ajax/embed-6/getSources",
        params={"id": url.name, "sId": sid},
    )

    sources = ajax_response.json()
    salt_secret = get_associative_key(session, SALT_SECRET_ENDPOINT).encode("utf-8")

    subtitles = [
        _.get("file") for _ in sources.get("tracks") if _.get("kind") == "captions"
    ]

    def yielder():
        for _ in json.loads(decipher_salted_aes(sources["sources"], salt_secret)):
            yield {
                "stream_url": _["file"],
                "subtitle": subtitles,
                "headers": {"SID": sid},
            }

        for _ in json.loads(decipher_salted_aes(sources["sourcesBackup"], salt_secret)):
            yield {
                "stream_url": _["file"],
                "subtitle": subtitles,
                "headers": {"SID": sid},
            }

    return list(yielder())



import functools
import json
from base64 import b64decode

import regex

from ...providers.nineanime import decipher

EMBED_URL_REGEX = regex.compile(r"(.+?/)(?:e(?:mbed)?)/([a-zA-Z0-9]+)")

SIMPLE_OPERATE_REGEX = regex.compile(r"[0-9]+\s+.+?\s+[0-9]+")


CIPHER_ALGORITHM_DATA_ENDPOINT = (
    "https://raw.githubusercontent.com/AnimeJeff/Overflow/main/syek"
)


def isolated_eval(code, *, builtins=None):
    try:
        return eval(code, {"__builtins__": builtins or {}}, {})
    except Exception:
        return None


@functools.lru_cache()
def get_ciper_algorithm(session):
    return json.loads(
        b64decode(
            b64decode(b64decode(session.get(CIPHER_ALGORITHM_DATA_ENDPOINT).content))
        ).decode()
    )


def string_encrypt(text):
    def repl(match):
        value = ord(match.group(0)) + 13
        return chr(
            value if (90 if ord(match.group(0)) <= 90 else 122) >= value else value - 26
        )

    return regex.sub(r"[a-zA-Z]", repl, text)


def conditional_encrypt(content, steps, b64_table):

    for _ in steps:
        if _ == "o":
            content = decipher.encrypt(content, table=b64_table).replace("/", "_")

        if _ == "s":
            content = string_encrypt(content)

        if _ == "a":
            content = content[::-1]

    return content


def extract(session, url, **opts):

    match = EMBED_URL_REGEX.search(url)
    algorithm = get_ciper_algorithm(session)

    b64_table = algorithm["encryptKey"]

    content_id = decipher.encrypt(
        decipher.cipher_keyed(algorithm["cipherKey"], match.group(2)),
        table=b64_table,
    )

    operations = algorithm["operations"]

    dashes = conditional_encrypt(
        "-".join(
            str(
                isolated_eval(
                    str(ord(character)) + operations[str(count % len(operations))]
                )
            )
            for count, character in enumerate(
                conditional_encrypt(content_id, algorithm["pre"], b64_table)
            )
        ),
        algorithm["post"],
        b64_table,
    )

    vidstream_info = session.get(f"{match.group(1)}/{algorithm['mainKey']}/{dashes}")

    return [
        {"stream_url": content.get("file", ""), "headers": {"referer": url}}
        for content in vidstream_info.json()["data"].get("media", {}).get("sources", [])
    ]



import logging

import click

from ...__version__ import __core__
from ...codebase import providers
from ...config import (
    CHECK_FOR_UPDATES,
    DEFAULT_PLAYER,
    DEFAULT_PROVIDER,
    DISCORD_PRESENCE,
    FORCE_STREAMING_QUALITY_SELECTION,
    QUALITY,
)
from .. import exit_codes, helpers, http_client

if DISCORD_PRESENCE:
    try:
        from ..helpers.rpc import set_streaming_episode
    except ImportError:
        raise ImportError(
            "Discord RPC was set to be enabled but `pypresence` is not installed, install it with `pip install pypresence`."
        )
    except RuntimeError:
        DISCORD_PRESENCE = False


@click.command(name="stream", help="Stream your favorite anime by query.")
@helpers.decorators.content_fetch_options(
    default_quality_string=QUALITY,
)
@helpers.decorators.player_options(default_player=DEFAULT_PLAYER)
@helpers.decorators.automatic_selection_options()
@helpers.decorators.logging_options()
@helpers.decorators.setup_loggers()
@helpers.decorators.banner_gift_wrapper(
    http_client.client, __core__, check_for_updates=CHECK_FOR_UPDATES
)
def animdl_stream(
    query, special, quality, player_opts, player, index, log_level, **kwargs
):
    """
    Streamer call for animdl streaming session.
    """
    r = kwargs.get("range")

    logger = logging.getLogger("streamer")
    streamer = helpers.handle_streamer(
        click.parser.split_arg_string(player_opts or "") or [], **{player: True}
    )

    if streamer is False:
        logger.critical(
            "Streaming failed due to selection of a unsupported streamer; please configure the streamer in the config to use it."
        )
        raise SystemExit(exit_codes.STREAMER_CONFIGURATION_REQUIRED)

    anime, provider = helpers.process_query(
        http_client.client, query, logger, auto_index=index, provider=DEFAULT_PROVIDER
    )

    if not anime:
        logger.critical("Searcher returned no anime to stream, failed to stream.")
        raise SystemExit(exit_codes.NO_CONTENT_FOUND)

    logger.name = "{}/{}".format(provider, logger.name)
    logger.debug("Will scrape from {}".format(anime))
    logger.info("Now initiating your stream session")

    match, provider_module, _ = providers.get_provider(
        providers.append_protocol(anime.get("anime_url"))
    )

    streams = list(
        provider_module.fetcher(http_client.client, anime.get("anime_url"), r, match)
    )

    if special:
        streams = list(helpers.special.special_parser(streams, special))

    content_title = anime["name"]

    total = len(streams)

    if total < 1:
        logger.critical("No streams found, failed to stream.")
        raise SystemExit(exit_codes.NO_CONTENT_FOUND)

    for count, (stream_urls_caller, episode_number) in enumerate(streams, 1):

        playing = True
        while playing:

            window_title = content_title + ": Episode {}".format(episode_number)

            stream_urls = list(
                helpers.ensure_extraction(http_client.client, stream_urls_caller)
            )

            if not stream_urls:
                logger.warning(
                    "There were no stream links available at the moment. Ignoring {!r}, retry using a different provider.".format(
                        window_title
                    )
                )
                playing = False
                continue

            selection = helpers.prompts.quality_prompt(
                logger,
                log_level,
                stream_urls,
                force_selection_string=quality
                if FORCE_STREAMING_QUALITY_SELECTION
                else None,
            )

            logger.debug("Calling streamer for {!r}".format(stream_urls))

            headers = selection.get("headers", {})
            _ = headers.pop("ssl_verification", True)

            logger.info(
                "Streaming {!r}, [{:d}/{:d}, {:d} remaining]".format(
                    window_title, count, total, total - count
                )
            )

            if "title" in selection:
                window_title += " ({})".format(selection["title"])

            player_process = streamer(
                selection.get("stream_url"),
                headers=headers,
                content_title=window_title,
                subtitles=selection.get("subtitle", []),
            )
            if DISCORD_PRESENCE:
                set_streaming_episode(http_client.client, content_title, episode_number)

            player_process.wait()

            if player_process.returncode:
                logger.warning(
                    "Detected a non-zero return code: {:d}.".format(
                        player_process.returncode
                    )
                )
                playing = (
                    False
                    if log_level > 20
                    else click.confirm(
                        "Retry playback for {!r}?".format(window_title),
                        show_default=True,
                        default=False,
                    )
                )
                continue

            playing = (
                False
                if log_level > 10
                else click.confirm(
                    "Replay {!r}?".format(window_title), show_default=True, default=True
                )
            )



"""
For when animixplay does not know
but the project knows.
"""

from functools import partial

import regex

REPL_REGEX = regex.compile(r"'(.+?)'")


hard_urls = {
    "overflow": {
        "provider": "bestcdn",
        "streams": [(f"Overflow/Overflow {_:02d}.mp4", _) for _ in range(1, 9)],
    },
    "jimihen-jimiko-wo-kaechau-jun-isei-kouyuu": {
        "provider": "bestcdn",
        "streams": [(f"Jimihen/Jimihen {_:02d}.mp4", _) for _ in range(1, 9)],
    },
}


def yield_from_bestcdn(session, attributes):

    repl_string = REPL_REGEX.search(
        session.get("https://anfruete.github.io/play/env.js").text
    ).group(1)

    for stream, episode in attributes["streams"]:
        yield partial(
            lambda stream: [{"stream_url": f"https://{repl_string}/{stream}"}], stream
        ), episode


yielder_mapping = {
    "bestcdn": yield_from_bestcdn,
}


def get_hardstream_generator(session, hard_stream):

    if hard_stream not in hard_urls:
        return

    attrs = hard_urls[hard_stream]
    provider = attrs["provider"]

    if provider not in yielder_mapping:
        return

    yielder = yielder_mapping[provider]
    return yielder(session, attrs)



from functools import partial

import lxml.html as htmlparser

from ....config import ANIMEONSEN
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(
    ANIMEONSEN,
    extra_regex=r"/(?:details/(?P<slug>[^?&/]+)|watch/(?P<watch_slug>[^?&/]+))",
)

API_ENDPOINT = "https://api.animeonsen.xyz/v4/"

AUTHENTICATION = "Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpcCI6IjI3LjM0LjY4LjEzOCIsImlhdCI6MTY0ODU1NTU4MiwiZXhwIjoxNjQ5MTYwMzgyLCJpc3MiOiJhbmltZW9uc2VuLWFwcCJ9.ZUy2zv_aRFtT_iqjEGtp5HO9EnkH71pM0TpcBQ0JcwU"


def get_stream_url(session, episode, slug):

    response = session.get(
        API_ENDPOINT + "content/{}/video/{}".format(slug, episode),
        headers={"Authorization": AUTHENTICATION},
    ).json()

    metadata = response.get("metadata", {})
    current, _, episodes = metadata.get("episode", [0, {}, {}])

    episode_title = episodes.get(str(current), {}).get(
        "contentTitle_episode_en"
    ) or episodes.get(str(current), {}).get("contentTitle_episode_jp", "")
    title = "Episode {}".format(current)

    if episode_title:
        title = "{}: {}".format(title, episode_title)

    streams = response.get("uri", {})

    return [
        {
            "stream_url": streams.get("stream"),
            "subtitles": list(streams.get("subtitles", {}).values()),
            "title": title,
            "headers": {
                "referer": ANIMEONSEN,
            },
        }
    ]


def fetcher(session, url, check, match):

    slug = match.group("slug") or match.group("watch_slug")

    for link in htmlparser.fromstring(
        session.get(ANIMEONSEN + "details/" + slug).text
    ).cssselect("div.episode-list > a"):
        episode_element = link.cssselect("div.episode")[0]

        episode = int(episode_element.get("data-episode", 0))

        if check(episode):
            yield partial(get_stream_url, session, episode, slug), episode


def metadata_fetcher(session, url, match):

    slug = match.group("slug") or match.group("watch_slug")

    response = session.get(
        API_ENDPOINT + "content/{}/extensive".format(slug),
        headers={"Authorization": AUTHENTICATION},
    ).json()

    titles = []

    if "content_title" in response:
        titles.append(response["content_title"])

    if "content_title_en" in response:
        titles.append(response["content_title_en"])

    return {"titles": titles}



from functools import partial

import regex

from ....config import ANIMTIME
from ...helper import construct_site_based_regex

CONTENT_RE = regex.compile(r"t\.(\w+)=([0-9]+)")
REGEX = construct_site_based_regex(ANIMTIME, extra_regex=r"/title/([^/?&]+)")


def get_content(url, js_content):
    content_slug = REGEX.match(url).group(1)
    for match in CONTENT_RE.finditer(js_content):
        if match.group(2) == content_slug:
            return match


def fetcher(session, url, check, match):

    content = session.get(
        ANIMTIME + "/main.fe2f7537e4a9d7929fe6.js", headers={"range": "bytes=386216-"}
    ).text

    anime = get_content(url, content)
    episodes = int(
        regex.search(
            r"zd.*?\[tm\.{}\]=(\d+)".format(regex.escape(anime.group(1))), content
        ).group(1)
    )
    constructor, end = regex.search(
        r'\[tm\.{}\]=function\(t\){{return"(.+?)"\+t\+"(.+?)"'.format(
            regex.escape(anime.group(1))
        ),
        content,
    ).groups()
    for episode in range(1, episodes + 1):
        if check(episode):
            yield partial(
                lambda x: [{"stream_url": x, "headers": {"referer": ANIMTIME}}],
                constructor + "{:03d}".format(episode) + end,
            ), episode


def metadata_fetcher(session, url, match):
    content = session.get(
        ANIMTIME + "/main.fe2f7537e4a9d7929fe6.js", headers={"range": "bytes=386216-"}
    ).text
    return {"titles": [get_content(url, content)]}



"""
A one cli for all the anime.
"""

import click

from .core import __version__
from .core.cli.commands import download, grab, schedule, search, stream, update
from .core.cli.helpers.player import supported_streamers

commands = {
    "download": download.animdl_download,
    "grab": grab.animdl_grab,
    "schedule": schedule.animdl_schedule,
    "search": search.animdl_search,
    "update": update.animdl_update,
}

executable = list(supported_streamers())

if executable:
    commands.update({"stream": stream.animdl_stream})


@click.group(commands=commands)
@click.version_option(__version__.__core__, "-v")
def __animdl_cli__():
    pass


if __name__ == "__main__":
    __animdl_cli__()



import json
import time

import regex

recv_data_regex = regex.compile(r"(\d+)(.+)")

POLLING_PARAMETERS = {
    "EIO": "4",
    "transport": "polling",
}


def parse_response(response_text):
    match = recv_data_regex.search(response_text)

    if match:
        return int(match.group(1)), json.loads(match.group(2))

    return None, response_text


def ws_stimulation(
    session,
    *,
    url="https://ws1.rapid-cloud.ru/socket.io/",
    close_event,
    sid_holder,
    parent_thread,
):
    def ws_poll(params={}, data=None):
        soft = POLLING_PARAMETERS.copy()
        soft.update(params)

        if data is None:
            response = session.get(url, params=soft)
        else:
            response = session.post(url, params=soft, data=data)

        return parse_response(response.text)

    _, response = ws_poll()

    ping_interval = response.get("pingInterval", 25000) // 1000
    polling_sid = response.get("sid")

    _, client_check = ws_poll(
        {
            "sid": polling_sid,
        },
        data="40",
    )

    if client_check != "ok":
        raise RuntimeError(
            f"Websocket server has returned a faulty value: {client_check!r}"
        )

    _, response = ws_poll(
        {
            "sid": polling_sid,
        }
    )

    user_sid = response.get("sid")

    sid_holder.update(sid=user_sid)

    while not active_sleep(
        ping_interval, lambda: not (close_event.is_set() or parent_thread.is_alive())
    ):

        _, data = ws_poll({"sid": polling_sid})

        if data != "2":
            raise RuntimeError(
                f"Websocket server has returned a faulty value: {data!r}"
            )

        _, data = ws_poll({"sid": polling_sid}, data="3")

        if data != "ok":
            raise RuntimeError(
                f"Websocket server has returned a faulty value: {data!r}"
            )


def active_sleep(interval, predicate, *, delay=0.1):

    time_now = time.time()

    while (time.time() - time_now) < interval:

        if predicate():
            return False

        time.sleep(delay)

    return True



from functools import partial

import lxml.html as htmlparser
import regex

from ....config import GOGOANIME
from ...helper import construct_site_based_regex

REGEX = construct_site_based_regex(
    GOGOANIME,
    extra_regex=r"/(?:(?P<episode_anime_slug>[^&?/]+)-episode-[\d-]+|category/(?P<anime_slug>[^&?/]+))",
)

ANIME_ID_REGEX = regex.compile(r'<input.+?value="(\d+)" id="movie_id"')
EPISODE_LOAD_AJAX = "https://ajax.gogo-load.com/ajax/load-list-episode"

TITLES_REGEX = regex.compile(r'<a href="/category/.+?" title=".+?">(.+?)</a>')


def get_episode_list(session, anime_id):
    """
    Fetch all the episodes' url from GogoAnime using.
    """
    episode_page = session.get(
        EPISODE_LOAD_AJAX,
        params={
            "ep_start": "0",
            "ep_end": "100000",
            "id": anime_id,
        },
    )
    genexp = iter(
        htmlparser.fromstring(episode_page.text).cssselect(
            'a[class=""] , a[class=""]  > div.name'
        )
    )

    for (a, div) in zip(genexp, genexp):
        episode = div.text_content().split(" ")[1]
        yield float(episode) if "." in episode else int(episode), GOGOANIME + a.get(
            "href"
        ).strip(),


def get_quality(url_text):
    match = regex.search(r"(\d+)P", url_text)
    if not match:
        return None
    return int(match.group(1))


def get_embed_page(session, episode_url):
    content_parsed = htmlparser.fromstring(session.get(episode_url).text)
    return "https:{}".format(content_parsed.cssselect("iframe")[0].get("src"))


def fetcher(session, url, check, match):

    if match.group("episode_anime_slug"):
        url = "{}category/{}".format(GOGOANIME, match.group("episode_anime_slug"))

    content_id = ANIME_ID_REGEX.search(session.get(url).text).group(1)

    for episode, episode_page in reversed(list(get_episode_list(session, content_id))):
        if check(episode):
            yield partial(
                lambda e: [
                    {
                        "stream_url": get_embed_page(session, e),
                        "further_extraction": ("gogoplay", {}),
                    }
                ],
                episode_page,
            ), episode


def metadata_fetcher(session, url, match):
    return {"titles": [TITLES_REGEX.search(session.get(url).text).group(1)]}



import shutil

import click
import regex
from anchor.strings import iter_contentaware_segments

from .constants import SOURCE_REPOSITORY, VERSION_FILE_PATH

terminal_columns = shutil.get_terminal_size((0, 0)).columns

VERSION_REGEX = regex.compile(r'__core__ = "(.*?)"')


def fetch_upstream_version(session):

    author, repository_name = SOURCE_REPOSITORY
    branch, version_file = VERSION_FILE_PATH

    upstream_version = VERSION_REGEX.search(
        session.get(
            f"https://raw.githubusercontent.com/{author}/{repository_name}/{branch}/{version_file}"
        ).text
    ).group(1)

    return upstream_version


def compare_version(
    current_version, upstream_version, columns=("major", "minor", "patch")
):

    if current_version == upstream_version:
        return False, None

    if current_version > upstream_version:
        return False, "Local version is more up-to-date than the upstream."

    return (
        True,
        f"{', '.join(f'{upstream_field - current_field} {field} change(s)' for field, current_field, upstream_field in zip(columns, map(int, current_version), map(int, upstream_version)) if upstream_field > current_field)} are available in the upstream.",
    )


def iter_banner(
    session,
    current_version,
    *,
    check_for_updates=True,
    description="A highly efficient, powerful and fast anime scraper.",
):

    author, repository_name = SOURCE_REPOSITORY

    if terminal_columns:
        processor = lambda _: iter_contentaware_segments(
            str.center(_, terminal_columns), terminal_columns
        )
    else:
        processor = lambda _: iter((_,))

    yield from processor(
        click.style(f"{author}/{repository_name} - v{current_version}", fg="magenta")
    )
    yield from processor(click.style(description, fg="magenta"))

    if check_for_updates:

        upstream_version = fetch_upstream_version(session)

        tuplised_upstream, tuplised_current_version = tuple(
            upstream_version.split(".")
        ), tuple(current_version.split("."))

        if tuplised_upstream > tuplised_current_version:
            yield from processor(
                click.style(
                    f"Update ↑ {upstream_version} ↓ {current_version}", fg="yellow"
                )
            )
            yield from processor(click.style(f"To update, `animdl update`."))


def banner_gift_wrapper(session, current_version, *, check_for_updates=False):
    def wrapper(f):
        def __inner__(*args, log_level, log_file, **kwargs):

            if log_level > 20:
                return f(*args, log_level=log_level, log_file=log_file, **kwargs)

            for _ in iter_banner(
                session,
                current_version,
                check_for_updates=check_for_updates,
            ):
                click.echo(_)

            return f(*args, log_level=log_level, log_file=log_file, **kwargs)

        return __inner__

    return wrapper



"""
All the search algorithms for all the providers available in AnimDL.
"""

import json

import lxml.html as htmlparser

from ...codebase.helper import uwu
from ...config import *
from .fuzzysearch import search

NINEANIME_URL_SEARCH = NINEANIME + "filter"

ANIMEPAHE_URL_CONTENT = ANIMEPAHE + "anime/%s"
ANIMEPAHE_URL_SEARCH_AJAX = ANIMEPAHE + "api"

ANIMEOUT_URL_SEARCH_AJAX = ANIMEOUT + "wp-admin/admin-ajax.php"

ANIMIX_URL_SEARCH_API = "https://cdn.animixplay.to/api/search"

ANIMIX_URL_SEARCH_POST = "https://v1.zv5vxk4uogwdp7jzbh6ku.workers.dev/"
ANIMIX_URL_CONTENT = ANIMIXPLAY.rstrip("/")

GOGOANIME_URL_SEARCH = GOGOANIME + "/search.html?"

TENSHI_URL_SEARCH_POST = TENSHI + "anime/search"
HAHO_URL_SEARCH_POST = HAHO + "anime/search"

TWIST_URL_CONTENT_API = "https://api.twist.moe/api/anime"
TWIST_URL_CONTENT = TWIST + "a/"


def placeholder(session, query):
    yield from []


def search_9anime(session, query):
    nineanime_results = session.get(NINEANIME_URL_SEARCH, params={"keyword": query})

    parsed = htmlparser.fromstring(nineanime_results.text)
    for results in parsed.cssselect("a.name"):
        yield {
            "anime_url": NINEANIME.rstrip("/") + results.get("href"),
            "name": results.text_content(),
        }


def search_animekaizoku(session, query):
    animekaizoku_results = htmlparser.fromstring(
        session.get(ANIMEKAIZOKU, params={"s": query}).text
    )
    for results in animekaizoku_results.cssselect(".post-title"):
        yield {
            "anime_url": ANIMEKAIZOKU + results.cssselect("a")[0].get("href"),
            "name": results.text_content(),
        }


def search_allanime(session, query):

    gql_response = session.get(
        ALLANIME + "graphql",
        params={
            "variables": '{"search":{"allowAdult":true,"query":"%s"},"translationType":"sub"}'
            % query.replace('"', '\\"'),
            "extensions": '{"persistedQuery":{"version":1,"sha256Hash":"9343797cc3d9e3f444e2d3b7db9a84d759b816a4d84512ea72d079f85bb96e98"}}',
        },
    )

    for result in gql_response.json().get("data", {}).get("shows", {}).get("edges", []):
        if any(a for k, a in result.get("availableEpisodes", {}).items()):
            yield {
                "anime_url": ALLANIME + "anime/{[_id]}".format(result),
                "name": result.get("name"),
            }


def search_animepahe(session, query):

    animepahe_results = session.get(
        ANIMEPAHE_URL_SEARCH_AJAX, params={"q": query, "m": "search"}
    )
    content = animepahe_results.json()

    for results in content.get("data", []):
        yield {
            "anime_url": ANIMEPAHE_URL_CONTENT % results.get("session"),
            "name": results.get("title"),
        }


def search_animeout(session, query):
    animeout_results = session.cf_request("GET", ANIMEOUT, params={"s": query})
    content = htmlparser.fromstring(animeout_results.text)

    for result in content.xpath('//h3[@class="post-title entry-title"]/a'):
        yield {"anime_url": result.get("href"), "name": result.text_content()}


def search_animixplay(session, query):
    animixplay_library = (
        _ for _ in session.get(ANIMIXPLAY + "assets/s/all.json").json() if _["e"] == "1"
    )

    for result in search(query, animixplay_library, processor=lambda r: r.get("title")):
        yield {
            "anime_url": ANIMIXPLAY + "v1/" + result.get("id"),
            "name": result.get("title"),
        }


def search_gogoanime(session, query):
    parsed = htmlparser.fromstring(
        session.get(GOGOANIME_URL_SEARCH, params={"keyword": query}).text
    )

    for results in parsed.xpath('//p[@class="name"]/a'):
        yield {
            "anime_url": GOGOANIME.strip("/") + results.get("href"),
            "name": results.get("title"),
        }


def search_kawaiifu(session, query):
    for results in htmlparser.fromstring(
        session.get(KAWAIIFU + "search-movie", params={"keyword": query}).text
    ).cssselect(".info > h4 > a:last-child"):
        yield {"anime_url": results.get("href"), "name": results.text_content().strip()}


def search_twist(session, query):
    content = session.get(
        TWIST_URL_CONTENT_API,
        headers={"x-access-token": "0df14814b9e590a1f26d3071a4ed7974"},
    )
    animes = content.json()

    for anime in search(
        query, animes, processor=lambda r: r.get("title") or r.get("alt_title")
    ):
        yield {
            "anime_url": TWIST_URL_CONTENT + anime.get("slug", {}).get("slug"),
            "name": anime.get("title", ""),
        }


def search_crunchyroll(session, query, *, scheme="http"):
    content = json.loads(
        session.get(
            CRUNCHYROLL + "ajax/?req=RpcApiSearch_GetSearchCandidates",
            headers={
                "Referer": "https://www.google.com/",
            },
        ).text.strip("*/\n -secur")
    )

    for anime in search(
        query, content.get("data", []), processor=lambda r: r.get("name")
    ):
        yield {
            "anime_url": scheme + anime.get("link", "").strip("/")[5:],
            "name": anime.get("name", ""),
        }


def search_nyaasi(session, query):
    for anime in htmlparser.fromstring(
        session.get(NYAASI, params={"q": query, "s": "seeders", "o": "desc"}).text
    ).cssselect('tr > td[colspan="2"] > a[title]:last-child	'):
        yield {
            "anime_url": NYAASI + anime.get("href")[1:],
            "name": anime.get("title", "").strip(),
        }


def search_tenshi(session, query, *, domain=TENSHI):
    uwu.bypass_ddos_guard(session, domain)
    tenshi_page = htmlparser.fromstring(
        session.get(domain + "anime", params={"q": query}).text
    )

    for result in tenshi_page.cssselect(".list > li > a"):
        yield {"name": result.get("title"), "anime_url": result.get("href")}


def search_zoro(session, query):
    for result in htmlparser.fromstring(
        session.get(ZORO + "/search", params={"keyword": query}).text
    ).cssselect("a.item-qtip[title][data-id]"):
        yield {
            "name": result.get("title"),
            "anime_url": ZORO + result.get("href")[1:-11],
        }


def search_h_ntai_stream(session, query):
    for result in htmlparser.fromstring(
        session.get(HENTAISTREAM, params={"s": query}).text
    ).cssselect("article > .bsx > a"):
        yield {
            "name": result.get("title"),
            "anime_url": result.get("href"),
        }


def search_haho(session, query):
    yield from search_tenshi(session, query, domain=HAHO)


provider_searcher_mapping = {
    "9anime": search_9anime,
    "animekaizoku": search_animekaizoku,
    "allanime": search_allanime,
    "animepahe": search_animepahe,
    "animeout": search_animeout,
    "animixplay": search_animixplay,
    "crunchyroll": search_crunchyroll,
    "kamyroll": lambda *args, **kwargs: search_crunchyroll(
        *args, **kwargs, scheme="kamyroll"
    ),
    "kawaiifu": search_kawaiifu,
    "gogoanime": search_gogoanime,
    "haho": search_haho,
    "hentaistream": search_h_ntai_stream,
    "tenshi": search_tenshi,
    "nyaa": search_nyaasi,
    "twist": search_twist,
    "zoro": search_zoro,
}



from typing import Callable, Iterable, TypeVar

import regex

search_type = TypeVar("search_type")


def search(
    query: "str",
    possibilities: "Iterable[search_type]",
    *,
    processor: "Callable[[search_type], str]" = lambda r: r
):

    pattern = regex.compile(
        r"(.*?)".join(map(regex.escape, query.strip())),
        flags=regex.IGNORECASE,
    )

    def genexp():
        for search_value in possibilities:
            match = pattern.search(processor(search_value))
            if match:
                yield len(search_value) - len(query.strip()), search_value

    for _, search_value in sorted(genexp(), key=lambda x: x[0]):
        yield search_value



from functools import partial

import lxml.html as htmlparser
import regex
import yarl

from ....config import ANIMEOUT
from ...helper import construct_site_based_regex, group_episodes, parse_from_content

REGEX = construct_site_based_regex(ANIMEOUT, extra_regex=r"/([^?&/]+)")

TITLES_REGEX = regex.compile(r'<h1 class="page-title">(.+?)</h1>')

public_domains = {
    "nimbus": "pub9",
    "chomusuke": "pub8",
    "chunchunmaru": "pub7",
    "ains": "pub6",
    "yotsuba": "pub5",
    "slaine": "pub4",
    "jibril": "pub3",
    "sv1": "pub2",
    "sv4": "pub2",
    "download": "pub1",
}


def animeout_stream_url(url: "yarl.URL") -> str:

    host_prefix, _ = url.host.split(".", 1)

    if host_prefix in public_domains:
        url = url.with_host(f"{public_domains[host_prefix]}.{_}")

    return url.human_repr()


def fetcher(session, url, check, match):
    animeout_page = session.cf_request("GET", url)
    parsed = htmlparser.fromstring(animeout_page.text)

    downloadables = (
        yarl.URL(_.get("href"))
        for _ in parsed.cssselect('.article-content a[href$="mkv"]')
        if "Download" in _.text_content()
    )

    for episode, content in sorted(
        (
            group_episodes(
                parse_from_content(
                    content,
                    name_processor=lambda c: c.name,
                    stream_url_processor=animeout_stream_url,
                )
                for content in downloadables
            ).items()
        ),
        key=lambda x: x[0],
    ):
        if check(episode):
            yield partial(list, content), episode


def metadata_fetcher(session, url, match):
    return {"titles": TITLES_REGEX.findall(session.cf_request("GET", url).text)}



import logging
import pathlib
import shutil
import subprocess
import sys

import click
import pip
import regex

from ...__version__ import __core__
from .. import helpers, http_client

GIT_CONFIG_REPO_NAME = regex.compile(
    r'\[remote ".+?"\].+?url = (\S+)', flags=regex.S
)


def is_repository(author, repository):

    if not pathlib.Path("./.git/").exists():
        return False

    repository_qualname = f"{author}/{repository}"

    with open("./.git/config") as git_config:
        match = GIT_CONFIG_REPO_NAME.search(git_config.read())

        if match is None:
            return False

    return match.group(1)[-len(repository_qualname) :] == repository_qualname


@click.command("update", help="Update the project or git pull if a repository.")
@helpers.decorators.setup_loggers()
def animdl_update(*args, **kwargs):

    if is_repository(*helpers.constants.SOURCE_REPOSITORY):
        update_logger = logging.getLogger(f"updater/git")

        if not shutil.which("git"):
            return update_logger.error("Couldn't find git to attempt.")

        process = subprocess.Popen(
            ["git", "pull"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    else:

        update_logger = logging.getLogger(f"updater/pip v{pip.__version__}")

        upstream_version = helpers.banner.fetch_upstream_version(http_client.client)
        tuplised_upstream, tuplised_current_version = tuple(
            upstream_version.split(".")
        ), tuple(__core__.split("."))

        comparison, description = helpers.banner.compare_version(
            tuplised_current_version, tuplised_upstream
        )

        if description is not None:
            update_logger.info(description)

        if not comparison:
            return update_logger.error("Already up to date.")

        update_logger.info(
            f"Updating module {helpers.constants.MODULE_NAME!r} via PIP: pip install --upgrade {helpers.constants.MODULE_NAME}"
        )

        process = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                helpers.constants.MODULE_NAME,
                "--user",
                "--no-warn-script-location",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

    for line in process.stdout:
        update_logger.info(f"[stdout] {line.decode().rstrip()}")

    for line in process.stderr:
        update_logger.error(f"[stderr] {line.decode().rstrip()}")



import lxml.html as htmlparser
import regex

STREAMLARE = "https://streamlare.com/"

CONTENT_ID = regex.compile(r"/[ve]/([^?#&/]+)")


def extract(session, url, **opts):

    csrf_token = (
        htmlparser.fromstring(session.get(url).text)
        .cssselect("meta[name='csrf-token']")[0]
        .get("content")
    )
    content_id = CONTENT_ID.search(url).group(1)

    def fast_yield():
        for _, streams in (
            session.post(
                STREAMLARE + "api/video/get",
                headers={
                    "x-requested-with": "XMLHttpRequest",
                    "x-csrf-token": csrf_token,
                },
                json={"id": content_id},
            )
            .json()
            .get("result")
            .items()
        ):
            yield {"stream_url": streams.get("src"), "headers": {"referer": STREAMLARE}}

    return list(fast_yield())



from base64 import b64decode
from hashlib import md5

from Cryptodome.Cipher import AES

TWISTMOE_SECRET = b"267041df55ca2b36f2e322d05ee2c9cf"

TWISTMOE_CDN = "https://{}cdn.twist.moe"
TWISTMOE_API = "https://api.twist.moe/api/anime/"


def unpad_content(content):
    return content[
        : -(content[-1] if isinstance(content[-1], int) else ord(content[-1]))
    ]


def generate_key(salt: bytes, *, output=48):

    key = md5(TWISTMOE_SECRET + salt).digest()
    current_key = key

    while len(current_key) < output:
        key = md5(key + TWISTMOE_SECRET + salt).digest()
        current_key += key

    return current_key[:output]


def decipher(encoded_url: str):

    s1 = b64decode(encoded_url.encode("utf-8"))
    assert s1.startswith(b"Salted__"), "Not a salt."
    key = generate_key(s1[8:16])
    return (
        unpad_content(AES.new(key[:32], AES.MODE_CBC, key[32:]).decrypt(s1[16:]))
        .decode("utf-8", "ignore")
        .lstrip(" ")
    )


def api(session, endpoint, content_slug):
    return session.get(
        TWISTMOE_API + content_slug + endpoint,
        headers={"x-access-token": "0df14814b9e590a1f26d3071a4ed7974"},
    )


def iter_episodes(session, content_slug):

    ongoing = api(session, "/", content_slug).json().get("ongoing")

    source_base = TWISTMOE_CDN.format("air-" if ongoing else "")

    episodes_page = api(session, "/sources", content_slug)

    if episodes_page.status_code >= 400:
        return

    for episode in api(session, "/sources", content_slug).json():
        yield episode.get("number", 0), source_base + decipher(episode.get("source"))



import base64
from functools import partial

import lxml.html as htmlparser
import regex

from ....config import HENTAISTREAM
from ...helper import construct_site_based_regex, uwu

EPISODE_REGEX = regex.compile(r"/\d+/[^&?/]+")
TITLES_REGEX = regex.compile(r'<h1 class="entry-title" itemprop="name">(.+?)</h1>')

REGEX = construct_site_based_regex(HENTAISTREAM, extra_regex=r"/(anime|\d+)/([^?&/]+)")

QUALITY_MAP = {
    720: "x264.720p.mp4",
    1080: "av1.1080p.webm",
    2160: "av1.2160p.webm",
    4320: "av1.4320p.webm",
}


def get_episodes_page(session, url):
    return (
        htmlparser.fromstring(session.get(url).text)
        .cssselect('li[itemscope] > a[href^="https://hentaistream.moe/anime/"]')[0]
        .get("href")
    )


def extract_from_site(session, episode_url, **opts):

    for embed in htmlparser.fromstring(session.get(episode_url).text).cssselect(
        "iframe"
    ):
        _, content_uri = embed.get("src", "#").split("#")
        base, *_sub = regex.split(r"[;,]", base64.b64decode(content_uri).decode()[4:])

        subtitle = list(base + _ + ".vtt" for _ in _sub)

        stream_meta = {**opts, **({"subtitle": subtitle} if subtitle else {})}

        for quality, suffix in QUALITY_MAP.items():
            stream = base + suffix
            if (
                session.head(stream, headers={"referer": HENTAISTREAM}).status_code
                == 200
            ):
                yield {
                    "quality": quality,
                    "stream_url": stream,
                    "headers": {"referer": HENTAISTREAM},
                    **stream_meta,
                }


def fetcher(session, url, check, match):
    uwu.bypass_ddos_guard(session, HENTAISTREAM)

    if match.group(1).isdigit():
        url = get_episodes_page(session, url)

    for episode_page in htmlparser.fromstring(session.get(url).text).cssselect(
        "li[data-index] > a"
    )[::-1]:
        number, title, date = (
            _.text_content() for _ in episode_page.cssselect('div[class^="epl"]')
        )
        episode_number = int(number) if number.isdigit() else 0

        if check(episode_number):
            yield partial(
                lambda ep, **opts: list(extract_from_site(session, ep, **opts)),
                episode_page.get("href"),
                title="{} ({})".format(title, date),
            ), episode_number


def metadata_fetcher(session, url, match):
    uwu.bypass_ddos_guard(session, HENTAISTREAM)

    if match.group(1).isdigit():
        url = get_episodes_page(session, url)

    return {"titles": TITLES_REGEX.findall(session.get(url).text)}



from base64 import b64decode
from hashlib import md5

from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import unpad


def generate_key_from_salt(salt: bytes, secret, *, output=48):

    key = md5(secret + salt).digest()
    current_key = key

    while len(current_key) < output:
        key = md5(key + secret + salt).digest()
        current_key += key

    return current_key[:output]


def decipher_salted_aes(encoded_url: str, secret, *, aes_mode=AES.MODE_CBC):

    raw_value = b64decode(encoded_url.encode("utf-8"))
    assert raw_value.startswith(b"Salted__"), "Not a salt."
    key = generate_key_from_salt(raw_value[8:16], secret)

    return (
        unpad(AES.new(key[:32], aes_mode, key[32:]).decrypt(raw_value[16:]), 16)
        .decode("utf-8", "ignore")
        .lstrip(" ")
    )
